import sympy as sy
import numpy as np
from numpy.linalg import norm
import math

def system_eq_func(f_, symbs_):
    funcs = []
    for i_f in f_:
        funcs.append(sy.lambdify(symbs_, i_f))
    return funcs


def system_eq_func_variables(f_, symbs_):
    funcs = []
    for i_f in f_:
        funcs.append(sy.lambdify(symbs_, i_f))
    return funcs


def deriv_eq(f_, symb_):
    return f_.diff(symb_)


def deriv_eq_variables(f_, symbs_):
    der_eq = []
    for i_syb in symbs_:
        der_eq.append(deriv_eq(f_, i_syb))
    return der_eq


def deriv_sys_variables(f_, symbs_):
    der_sys = []
    for i_f in f_:
        der_sys.append(deriv_eq_variables(i_f, symbs_))
    return der_sys


def get_symbols_from_sys(symbs_):
    return symbs_


def Jacobian_matrix(f_, symbs_):
    jac_mat = []
    symbols_ = get_symbols_from_sys(symbs_)
    d_s_v = deriv_sys_variables(f_, symbols_)
    for i_s in d_s_v:
        i_m = system_eq_func_variables(i_s, symbols_)
        jac_mat.append(i_m)
    return jac_mat


def Jacobian_matrix_values(jac_, symb_values):
    jac_mat_v = []
    # d_s_v = deriv_sys_variables(f_, symbols_)
    for i_s in jac_:
        i_jac_mat_v = []
        for j_s in i_s:
            val_m_i_j = j_s(*symb_values)
            i_jac_mat_v.append(val_m_i_j)
        jac_mat_v.append(i_jac_mat_v)
    return jac_mat_v


def lambda_h(jac_, system_):
    J_p = Jacobian_matrix_values(jac_, system_)
    norm_q = 1 / norm(system_) ** 2
    return norm_q * (np.dot(system_, J_p).dot(np.array(system_).T))


def euler_d_system(X_, s_, w_, dt_, guard_equations_, reset_equations_):
    dX_ = []
    for index_s in range(0, len(s_)):
        # print("s_=", s_[index_s](*X_))
        # print("w_=", w_)
        # print("dt_=", dt_)
        dX_.append(X_[index_s] + dt_ * (s_[index_s](*X_) + w_))
    if guard_equations_:
        if guard_equations_[0](*X_):
            print("g_=")
            for index_s in range(0, len(s_)):
                if reset_equations_[index_s]:
                    dX_[index_s] = reset_equations_[index_s](*X_)
    return dX_


def dis_2_pts_1D(point1, point2):
    return abs(point1 - point2)


def gamma_u(system_1, system_2, d_system_1, d_system_2, lambda_u, w_):
    diff_d_system = [x - y for x, y in zip(d_system_1, d_system_2)]
    diff_system = [x - y for x, y in zip(system_1, system_2)]
    diff_w = 2 * w_

    # print("norm(diff_w)=", norm(diff_w))
    t1 = diff_d_system
    t2 = diff_system
    a1 = np.dot(t1, t2)
    # print("diff_system=", diff_system, "diff_d_system=", diff_d_system, "scalair product=", a1, "  lambda_u*(norm(diff_system)**2)=", lambda_u*(norm(diff_system)**2), "  norm(diff_system)=", norm(diff_system), "   norm(diff_w)=", norm(diff_w))
    return (a1 - lambda_u * (norm(diff_system) ** 2)) / (norm(diff_system) * norm(diff_w))


def all_gamma_u(all_system_, all_system_next, lambda_u, w_):
    sup_gamma_u = -500000
    # print("all_system_=", all_system_, "all_system_next=", all_system_next, "lambda=", lambda_u)
    # print("len(all_system_)=", len(all_system_), "len(all_system_next)=", len(all_system_next), "lambda=", lambda_u)
    for i_all in range(0, len(all_system_)):
        for j_all in range(0, len(all_system_)):
            # print(" all_w[i_all] != all_w[j_all]", "i_all = ", i_all, "j_all", j_all)
            if i_all != j_all:
                d_system_i = all_system_next[i_all]
                d_system_j = all_system_next[j_all]
                gamma_u_cur = gamma_u(all_system_[i_all], all_system_[j_all], d_system_i,
                                      d_system_j, lambda_u, w_)
                # print("gamma_u_cur=", gamma_u_cur, "   j_all = ", j_all, "        sup_gamma_u=", sup_gamma_u)
                if sup_gamma_u < gamma_u_cur:
                    sup_gamma_u = gamma_u_cur
                    # print("sup_gamma_u=", sup_gamma_u, "   j_all = ", j_all)
    return sup_gamma_u


def Lu(system_1, system_2, d_system_1, d_system_2):
    diff_d_bio = [x - y for x, y in zip(d_system_1, d_system_2)]
    diff_bio = [x - y for x, y in zip(system_1, system_2)]
    return norm(diff_d_bio) / norm(diff_bio)
    # return math.sqrt((d_system_1[0]-d_system_2[0])**2 + (d_system_1[1]-d_system_2[1])**2 + (d_system_1[2]-d_system_2[2])**2) / math.sqrt((system_1[0]-system_2[0])**2 + (system_1[1]-system_2[1])**2 + (system_1[2]-system_2[2])**2)


def Lu_all(all_system_):
    sup_Lu = -5000000
    for i_all in range(0, len(all_system_) - 1):
        for j_all in range(0, len(all_system_) - 1):
            if i_all != j_all:
                d_system_i = all_system_[i_all + 1]
                d_system_j = all_system_[j_all + 1]
                # print(all_system_[i_all], all_system_[j_all], d_system_i, d_system_j)
                Lu_cur = Lu(all_system_[i_all], all_system_[j_all], d_system_i, d_system_j)
                # print(Lu_cur)
                if sup_Lu < Lu_cur:
                    sup_Lu = Lu_cur
    return sup_Lu


def Cu_all(Lu_, all_system_):
    sup_Cu = -5000000
    for i_all in range(1, len(all_system_)):
        Cu_cur = Lu_ * norm(all_system_[i_all])
        if Cu_cur > sup_Cu:
            sup_Cu = Cu_cur
    return sup_Cu

def delta_t_with_uncertainty_lambda_neg(Cu_, lambda_u_, gamma_u_, epsilon_, W_, t_):
    # print(norm(W_))
    l1 = (Cu_ ** 2 / (-lambda_u_ ** 4)) * (
            -(lambda_u_ ** 2) * (t_ ** 2) - 2 * lambda_u_ * t_ + 2 * math.exp(lambda_u_ * t_) - 2)
    l2 = (1 / lambda_u_ ** 2) * (
            ((Cu_ * gamma_u_ * (norm(W_))) / (-lambda_u_)) * (-lambda_u_ * t_ + math.exp(lambda_u_ * t_) - 1))
    l3 = (1 / lambda_u_) * (((gamma_u_ ** 2 * (norm(W_) / 2) ** 2) / (-lambda_u_)) * (math.exp(lambda_u_ * t_) - 1) + (
            lambda_u_ * (epsilon_ ** 2) * math.exp(lambda_u_ * t_)))
    return math.sqrt(l1 + l2 + l3)

def delta_t_without_uncertainty_lambda_neg(Cu_, lambda_u_, epsilon_, t_):
    l1 = (Cu_ ** 2 / (-lambda_u_ ** 4)) * (
            -lambda_u_ ** 2 * t_ ** 2 - 2 * lambda_u_ * t_ + 2 * math.exp(lambda_u_ * t_) - 2)
    # l2 = (1/lambda_u_**2)*(((Cu_*gamma_u_*norm(W_))/(-lambda_u_)) * (-lambda_u_*t_ + math.exp(lambda_u_*t_) - 1))
    l3 = (1 / lambda_u_) * (lambda_u_ * epsilon_ ** 2 * math.exp(lambda_u_ * t_))
    return math.sqrt(l1 + l3)

def delta_t_without_uncertainty_lambda_pos(Cu_, lambda_u_, epsilon_, t_):
    l1 = (epsilon_ ** 2) * math.exp(3 * lambda_u_ * t_)
    l2 = (Cu_ ** 2 / (3 * lambda_u_ ** 2)) * (
            -t_ ** 2 - (2 * t_ / (3 * lambda_u_)) + (2 / (9 * lambda_u_ ** 2)) * (math.exp(3 * lambda_u_ * t_) - 1))
    return math.sqrt(l1 + l2)

def delta_t_with_uncertainty_lambda_pos(Cu_, lambda_u_, gamma_u_, epsilon_, W_, t_):
    coef_ = 1 / ((3 * lambda_u_) ** (3 / 2))
    l1 = (Cu_ ** 2 / lambda_u_) * (
            -9 * (lambda_u_ ** 2) * (t_ ** 2) - 6 * lambda_u_ * t_ + 2 * math.exp(3 * lambda_u_ * t_) - 2)
    l2 = 3 * lambda_u_ * (
            ((Cu_ * gamma_u_ * norm(W_)) / lambda_u_) * (-3 * lambda_u_ * t_ + math.exp(3 * lambda_u_ * t_) - 1))
    l3 = (9 * lambda_u_ ** 2) * (
            (((gamma_u_ ** 2) * ((norm(W_) / 2) ** 2)) / lambda_u_) * (math.exp(3 * lambda_u_ * t_) - 1) + (
            3 * lambda_u_ * epsilon_ ** 2 * math.exp(3 * lambda_u_ * t_)))
    return coef_ * math.sqrt(l1 + l2 + l3)

def delta_t_with_uncertainty_lambda_null(Cu_, gamma_u_, epsilon_, W_, t_):
    l1_ = (Cu_**2)*(- (t_ ** 2) - 2 * t_ + 2 * math.exp(t_) - 2)
    l2_ = Cu_*gamma_u_*abs(W_)*(-t_ + math.exp(t_)-1) + ((gamma_u_**2) * ((abs(W_)/2)**2) * (math.exp(t_) - 1) + epsilon_ ** 2 * math.exp(t_))
    # d_ = ((Cu_**2)*(-(t_)**2 - 2*t_ + 2*math.exp(t_) - 2) + (Cu_ * gamma_u_*abs(W_)* (-t_ + math.exp(t_)-1 ) + (gamma_u_**2 * (abs(W_) / 2) ** 2 * (math.exp(t_) - 1) + epsilon_ ** 2 * math.exp(t_))))**(1 / 2)
    return math.sqrt(l1_ + l2_)

def delta_t_without_uncertainty_lambda_null(Cu_, epsilon_, t_):
    l1_ = (Cu_**2)*(- (t_ ** 2) - 2 * t_ + 2 * math.exp(t_) - 2)
    l2_ = epsilon_ ** 2 * math.exp(t_)
    # d_ = ((Cu_**2)*(-(t_)**2 - 2*t_ + 2*math.exp(t_) - 2) + (Cu_ * gamma_u_*abs(W_)* (-t_ + math.exp(t_)-1 ) + (gamma_u_**2 * (abs(W_) / 2) ** 2 * (math.exp(t_) - 1) + epsilon_ ** 2 * math.exp(t_))))**(1 / 2)
    return math.sqrt(l1_ + l2_)

def get_states_name_string(states_):
    return [str(states_i_).replace("_sy", "") for states_i_ in states_]

def index_sys_lambda_sign_change(system_, lambda_):
    index_sys = [0]
    first_sign_positive = True if lambda_[0] > 0 else False
    for i_com in range(1, len(system_)):
        if lambda_[i_com - 1] * lambda_[i_com] < 0:
            index_sys.append(i_com)
    index_sys.append(len(system_) - 1)
    return index_sys, first_sign_positive

def find_period(init_system_, num_periods_, system_equations_, dt_, guard_equations_, reset_equations_):
    prev_prev_system_period = init_system_
    prev_system_period = euler_d_system(prev_prev_system_period, system_equations_, 0, dt_, guard_equations_, reset_equations_)
    i_period = 0
    d_system_period = euler_d_system(prev_system_period, system_equations_, 0, dt_, guard_equations_, reset_equations_)
    i_period_1 = []
    while len(i_period_1) <= num_periods_:
        cond_priodicity = (prev_prev_system_period[1] < prev_system_period[1]) and (prev_system_period[1] > d_system_period[1])
        if cond_priodicity:
            print(i_period)
            i_period_1.append(i_period + 1)

        prev_prev_system_period = prev_system_period
        prev_system_period = d_system_period
        d_system_period = euler_d_system(d_system_period, system_equations_, 0, dt_, guard_equations_, reset_equations_)
        i_period += 1
    i_one_period_ = i_period_1[num_periods_ - 1] - i_period_1[num_periods_ - 2] + 1
    return i_one_period_, d_system_period

